/*
Copyright (c) 2022 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#if defined(_MSC_VER) && !defined(__clang__)
#  pragma warning(push)
#  pragma warning(disable : 4706 4711) // suppress warnings due to assignment within conditional expression, and automatically inlined functions
#  if _MSC_VER >= 1914
#    pragma warning(disable : 5045) // suppress information that MSVC will insert Spectre mitigation if /Qspectre switch specified
#  endif
#endif

#include "findinfile.h"
#include <limits.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#ifdef _WIN32
#  include <share.h> // definition of _SH_DENYNO
#  define fopen_(_filename, _mode)         _fsopen(_filename, _mode, _SH_DENYNO) // fopen is marked both unsafe and deprecated for Windows
#  define wfopen_(_filename, _mode)        _wfsopen(_filename, _mode, _SH_DENYNO)
#  define ftell_(_file)                    _ftelli64(_file)                   // usually implementations on Windows x64 use data model LLP64, ...
#  define fseek_(_file, _offset, _origin)  _fseeki64(_file, _offset, _origin) // ... so we would still stuck with a 32 bit 'long' type without this branch
typedef int64_t long_t_;
#else // ! _WIN32
#  define fopen_(_filename, _mode)         fopen(_filename, _mode)
#  define wfopen_(_filename, _mode)        NULL
#  define ftell_(_file)                    ftell(_file)
#  define fseek_(_file, _offset, _origin)  fseek(_file, _offset, _origin)
typedef long long_t_;
#endif // ! _WIN32
typedef unsigned char byte_t_;

// Help rounding upward to the next power of two that is not less than n. Relies on n > 0.
static size_t ceil_pow_two_(const long n)
{
  register size_t c = (const size_t)n - 1U;
  c |= c >> 1U;
  c |= c >> 2U;
  c |= c >> 4U;
  c |= c >> 8U;
  c |= c >> 16U;
#if (ULONG_MAX > 4294967295UL) // most 64 bit platforms but Windows
  c |= c >> 32U;
#endif
  return c + 1U;
}

// The following two private functions implement the Rabin-Karp search algorithm.
static FILE *ffind_search_(FILE *const file_stream, const void *const needle, const long needle_cnt)
{
  // BUFSIZ is likely the number of bytes used by the internal stream buffer and thus, it's the size we prefer to read at once.
  // If the length of the needle sequence is greater than BUFSIZ, we would get overlapped memory of the preceded bytes in our buffer,
  // and the position from where we copied these bytes. We round needle_cnt up to the next power of two in this case.
  const size_t chunk_max = (needle_cnt <= BUFSIZ) ? BUFSIZ : ceil_pow_two_(needle_cnt);

  // The total buffer size consists of preceded bytes where we copy the last needle_cnt bytes of the previous chunk (if any), and the chunk size in which we read new bytes from the file.
  byte_t_ *const buffer = (byte_t_ *)malloc((const size_t)needle_cnt + chunk_max);
  if (!buffer)
    return NULL;

  byte_t_ *const chunk_start = buffer + needle_cnt; // Position in the buffer which is the pointer where the newly read chunk begins.
  size_t read_cnt = fread(chunk_start, 1U, chunk_max, file_stream); // Read the first chunk and capture the length that have been read.
  if (read_cnt < (const size_t)needle_cnt) // If the file content is shorter than needle it will be meaningless to begin searching.
  {
    free(buffer);
    return NULL;
  }

  static const long hash_bitcnt = (long)sizeof(uintmax_t) * CHAR_BIT; // size of the hashes in bits (constant expression at compile time)
  const int is_long_needle = (needle_cnt > hash_bitcnt); // 1 if the number of bytes in needle exceeds the the number of bits of the hash type, 0 otherwise
  const uintmax_t eliminator = is_long_needle ? (uintmax_t)0U : ((uintmax_t)1U << (const unsigned long)(needle_cnt - 1L)); // factor used to remove bytes from chunk_hash
  uintmax_t chunk_hash = 0U, // rolling hash of the buffered bytes of the file
    needle_hash = 0U; // rolling hash of the bytes in needle
  const byte_t_ *seq_cursor = chunk_start, // points to the buffered byte were the sequence begins, which will always be (chunk_cursor - needle_cnt) later on
    *chunk_cursor = is_long_needle ? chunk_start + ((long_t_)needle_cnt - hash_bitcnt) : chunk_start, // points to the buffered byte which we are about to process
    *chunk_end = chunk_start + read_cnt, // address which marks the end of the chunk read
    *const cpy_ptr = chunk_end - needle_cnt; // address from where to copy the last needle_cnt bytes if we have to read the next chunk

  // calculate both the hash of needle and the initial hash of the buffered bytes
  for (const byte_t_ *const end = (const byte_t_ *)needle + needle_cnt, *needle_cursor = is_long_needle ? end - hash_bitcnt : (const byte_t_ *)needle; needle_cursor < end;)
  {
    needle_hash = (needle_hash << 1U) + *needle_cursor++;
    chunk_hash = (chunk_hash << 1U) + *chunk_cursor++;
  }

  for (;;) // Comparison and searching begins here.
  {
    // Same hashes indicate that we found the sequence. However, we still have to compare the whole sequence in this case.
    if (needle_hash == chunk_hash && memcmp(seq_cursor, needle, (const size_t)needle_cnt) == 0)
    {
      if (fseek_(file_stream, (long_t_)(seq_cursor - chunk_end), SEEK_CUR) != 0) // we need to move the stream cursor back to the begin of the found sequence
        break;

      free(buffer);
      return file_stream;
    }

    // If all bytes of the chunk are processed, we need to read more data.
    if (chunk_cursor == chunk_end)
    {
      if (read_cnt < chunk_max) // we leave the function if we already reached the end of the file when we read the latest chunk
        break;

      seq_cursor = (const byte_t_ *)memcpy(buffer, cpy_ptr, (const size_t)needle_cnt); // copy the last needle_cnt bytes to the begin of the buffer, also reset the cursor for the sequence
      if ((read_cnt = fread(chunk_start, 1U, chunk_max, file_stream)) == 0U) // try to read the next chunk; leave the function if no further bytes have been read
        break;

      chunk_cursor = chunk_start; // reset the cursor for the buffered bytes
      chunk_end = chunk_start + read_cnt; // update the end position of the chunk
    }

    chunk_hash = ((chunk_hash - *seq_cursor++ * eliminator) << 1U) + *chunk_cursor++; // we update the hash of the buffered bytes
  }

  free(buffer);
  return NULL;
}

static FILE *frfind_search_(FILE *const file_stream, const void *const needle, const long needle_cnt)
{
  long_t_ remaining_file_size = ftell_(file_stream); // Current position of the file cursor.
  if (remaining_file_size < needle_cnt) // If the remaining file content is shorter than needle it will be meaningless to begin searching.
    return NULL;

  const size_t chunk_max = (needle_cnt <= BUFSIZ) ? BUFSIZ : ceil_pow_two_(needle_cnt); // If necessary, round needle_cnt up to the next power of two.

  // The total buffer size consists of the chunk size in which we read new bytes from the file, and appended bytes where we copy the last needle_cnt bytes of the previous chunk (if any).
  byte_t_ *const buffer = (byte_t_ *)malloc(chunk_max + (const size_t)needle_cnt);
  if (!buffer)
    return NULL;

  byte_t_ *const appendix_start = buffer + chunk_max, // Position in the buffer to where the last needle_cnt bytes of the previous chunk are copied.
    *chunk_start = (size_t)remaining_file_size < chunk_max ? appendix_start - remaining_file_size : buffer; // Position in the buffer which is the pointer where the newly read chunk begins.
  long_t_ to_read = (long_t_)(appendix_start - chunk_start); // Number of bytes to read.
  if (fseek_(file_stream, -to_read, SEEK_CUR) != 0) // Move the file cursor to the position where reading begins.
  {
    free(buffer);
    return NULL;
  }

  long_t_ read_cnt = (long_t_)fread(chunk_start, 1U, (size_t)to_read, file_stream); // Read the first chunk and capture the length that have been read.
  if (read_cnt != to_read)
  {
    free(buffer);
    return NULL;
  }

  remaining_file_size -= read_cnt;
  static const long hash_bitcnt = (long)sizeof(uintmax_t) * CHAR_BIT; // size of the hashes in bits (constant expression at compile time)
  const int is_long_needle = (needle_cnt > hash_bitcnt); // 1 if the number of bytes in needle exceeds the the number of bits of the hash type, 0 otherwise
  const uintmax_t eliminator = is_long_needle ? (uintmax_t)0U : ((uintmax_t)1U << (const unsigned long)(needle_cnt - 1L)); // factor used to remove bytes from chunk_hash
  uintmax_t chunk_hash = 0U, // rolling hash of the buffered bytes of the file
    needle_hash = 0U; // rolling hash of the bytes in needle

  const byte_t_ *seq_cursor = appendix_start, // points to past-the-current buffered byte were the sequence ends, which will always be (appendix_start + needle_cnt) later on
    *chunk_cursor = is_long_needle ? appendix_start - needle_cnt + hash_bitcnt : appendix_start; // points to past-the-current buffered byte which we are about to process

  // calculate both the hash of needle and the initial hash of the buffered bytes
  for (const byte_t_ *const rend = (const byte_t_ *)needle, *needle_cursor = is_long_needle ? rend + hash_bitcnt : rend + needle_cnt; needle_cursor-- > rend;)
  {
    needle_hash = (needle_hash << 1U) + *needle_cursor;
    chunk_hash = (chunk_hash << 1U) + *--chunk_cursor;
  }

  for (;;) // Comparison and searching begins here.
  {
    // Same hashes indicate that we found the sequence. However, we still have to compare the whole sequence in this case.
    if (needle_hash == chunk_hash && memcmp(chunk_cursor, needle, (const size_t)needle_cnt) == 0)
    {
      if (fseek_(file_stream, (long_t_)(chunk_cursor - appendix_start), SEEK_CUR) != 0) // we need to move the stream cursor back to the begin of the found sequence
        break;

      free(buffer);
      return file_stream;
    }

    // If all bytes of the chunk are processed, we need to read more data.
    if (chunk_cursor == chunk_start)
    {
      if (!remaining_file_size) // we leave the function if we already reached the begin of the file when we read the latest chunk
        break;

      memcpy(appendix_start, chunk_start, (const size_t)needle_cnt); // copy the last needle_cnt bytes to the appendix
      chunk_start = (size_t)remaining_file_size < chunk_max ? appendix_start - remaining_file_size : buffer;
      to_read = (long_t_)(appendix_start - chunk_start);
      if (fseek_(file_stream, -(read_cnt + to_read), SEEK_CUR) != 0 || (read_cnt = (long_t_)fread(chunk_start, 1U, (size_t)to_read, file_stream)) != to_read) // move back and read the next chunk
        break;

      remaining_file_size -= read_cnt;
      seq_cursor = appendix_start + needle_cnt;
      chunk_cursor = appendix_start; // reset the cursor for the buffered bytes
    }

    chunk_hash = ((chunk_hash - *--seq_cursor * eliminator) << 1U) + *--chunk_cursor; // we update the hash of the buffered bytes
  }

  free(buffer);
  return NULL;
}

// Code shared with all public functions (ffind, ffindw, frfind, and frfindw).
static FILE *ffind_generic_(const void *const file_name, const int is_wide, FILE *const file_stream, const void *const needle, const long needle_cnt, const int is_reverse)
{
  if (!needle || needle_cnt < 1L)
    return NULL;

  long_t_ initial_pos = 0; // captures the current position of the stream cursor
  FILE *stream = NULL;
  if (file_stream)
  {
    if (fseek_(file_stream, 0, SEEK_CUR) != 0) // ensure we begin reading at the current position of the stream cursor (if the last action was a write operation)
      return NULL;

    stream = file_stream;
    initial_pos = ftell_(stream);
  }
  else if (!file_name || !(stream = !is_wide ? fopen_((const char *const)file_name, "r+b") : wfopen_((const wchar_t *const)file_name, L"r+b")) || // narrow-/wide-string branching
           (is_reverse && fseek_(stream, 0, SEEK_END) != 0)) // set the file pointer to the end of the file if we have to search backwards
    return NULL;

  FILE *ret = !is_reverse ? ffind_search_(stream, needle, needle_cnt) : frfind_search_(stream, needle, needle_cnt); // search implementations
  if (ret)
    return ret;

  if (file_stream)
    fseek_(stream, initial_pos, SEEK_SET); // reset the stream cursor if we didn't find the sequence
  else
    fclose(stream);

  return NULL;
}

FILE *ffind(const char *file_name, FILE *file_stream, const void *needle, long needle_cnt)
{
  return ffind_generic_((const void *const)file_name, 0, file_stream, needle, needle_cnt, 0);
}

FILE *frfind(const char *file_name, FILE *file_stream, const void *needle, long needle_cnt)
{
  return ffind_generic_((const void *const)file_name, 0, file_stream, needle, needle_cnt, 1);
}

#ifdef _WIN32
FILE *ffindw(const wchar_t *file_name, FILE *file_stream, const void *needle, long needle_cnt)
{
  return ffind_generic_((const void *const)file_name, 1, file_stream, needle, needle_cnt, 0);
}

FILE *frfindw(const wchar_t *file_name, FILE *file_stream, const void *needle, long needle_cnt)
{
  return ffind_generic_((const void *const)file_name, 1, file_stream, needle, needle_cnt, 1);
}
#endif // _WIN32

#if defined(_MSC_VER) && !defined(__clang__)
#  pragma warning(pop)
#endif
