/*
Copyright (c) 2022 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
/// @file      findinfile.h
/// @brief     C interface to find byte sequences in a file.
/// @version   1.2
/// @author    Steffen Illhardt
/// @date      2022
/// @copyright MIT License
/// @pre       Requires compiler support for at least C99 (and platform-specific extensions).

#ifndef FINDINFILE_H_INCLUDED__
#define FINDINFILE_H_INCLUDED__

#ifndef __cplusplus
#  include <stddef.h>
#  include <stdio.h>
#else // __cplusplus
#  include <cstdio>
extern "C" {
#endif // __cplusplus

/// @brief  Search a byte sequence in a binary file. <br>
///          This function implements the Rabin-Karp algorithm.
/// @param  file_name    Name of the file. This parameter may be `NULL` if the
///                       `file_stream` parameter is not `NULL`. Searching
///                       starts at the begin of the file.
/// @param  file_stream  File stream previously opened using fopen in binary
///                       mode with at least read access, or `NULL` if
///                       `file_name` is specified. If this parameter is not
///                       `NULL`, the `file_name` parameter will be ignored.
///                       Searching starts at the current position of the stream
///                       cursor. The stream cursor is set to the begin of the
///                       found sequence if the function succeeds.
/// @param  needle       Byte sequence to be searched within the file content.
/// @param  needle_cnt   Number of bytes to be used from the begin of the
///                       `needle` parameter.
/// @return One of the following: <br>
///         - Pointer to `FILE` with the stream cursor set to the begin of the
///            found sequence. If `file_stream` is `NULL` it will point to a
///            stream opened in "r+b" mode. <br>
///         - `NULL` if the function failed. If `file_stream` is not `NULL`, the
///            stream is still open, and the stream cursor of `file_stream` is
///            pointing to the same position when the function was called.
FILE *ffind(const char *file_name, FILE *file_stream, const void *needle, long needle_cnt);

#ifdef _WIN32
/// @brief  Search a byte sequence in a binary file (Windows only). <br>
///         Prefer to use this function on Windows for Unicode support of file names. <br>
///         For further information refer to the description of `ffind()`.
FILE *ffindw(const wchar_t *file_name, FILE *file_stream, const void *needle, long needle_cnt);
#endif // _WIN32

/// @brief  Search a byte sequence in a binary file in reverse direction. <br>
///          This function implements the Rabin-Karp algorithm.
/// @param  file_name    Name of the file. This parameter may be `NULL` if the
///                       `file_stream` parameter is not `NULL`. Searching
///                       starts at the end of the file.
/// @param  file_stream  File stream previously opened using fopen in binary
///                       mode with at least read access, or `NULL` if
///                       `file_name` is specified. If this parameter is not
///                       `NULL`, the `file_name` parameter will be ignored.
///                       Backward searching starts at the current position of
///                       the stream cursor. The stream cursor is set to the
///                       begin of the found sequence if the function succeeds.
/// @param  needle       Byte sequence to be searched within the file content.
/// @param  needle_cnt   Number of bytes to be used from the begin of the
///                       `needle` parameter.
/// @return One of the following: <br>
///         - Pointer to `FILE` with the stream cursor set to the begin of the
///            found sequence. If `file_stream` is `NULL` it will point to a
///            stream opened in "r+b" mode. <br>
///         - `NULL` if the function failed. If `file_stream` is not `NULL`, the
///            stream is still open, and the stream cursor of `file_stream` is
///            pointing to the same position when the function was called.
FILE *frfind(const char *file_name, FILE *file_stream, const void *needle, long needle_cnt);

#ifdef _WIN32
/// @brief  Search a byte sequence in a binary file in reverse direction (Windows only). <br>
///         Prefer to use this function on Windows for Unicode support of file names. <br>
///         For further information refer to the description of `frfind()`.
FILE *frfindw(const wchar_t *file_name, FILE *file_stream, const void *needle, long needle_cnt);
#endif // _WIN32

#ifdef __cplusplus
}
#endif // __cplusplus

#endif // FINDINFILE_H_INCLUDED__
